/**
 * @file boost/mirror/meta_namespace.hpp
 * @brief Implementation of namespace registering and reflection
 *
 *
 *  Copyright 2008-2010 Matus Chochlik. Distributed under the Boost
 *  Software License, Version 1.0. (See accompanying file
 *  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef BOOST_MIRROR_META_NAMESPACE_1011291729_HPP
#define BOOST_MIRROR_META_NAMESPACE_1011291729_HPP

#include <boost/mirror/mirror_fwd.hpp>
#include <boost/mirror/tag_dispatch.hpp>
#include <boost/mirror/auxiliary/scoped_named.hpp>
#include <boost/mirror/auxiliary/global_list.hpp>
#include <string>
#include <type_traits>

BOOST_MIRROR_NAMESPACE_BEGIN

template <class Namespace>
struct meta_namespace
 : public aux::scoped_named_impl<Namespace>
{
        typedef meta_namespace<typename Namespace::scope> scope;
};

// Implementation of the members meta-function for namespaces
template <class Namespace>
struct members<meta_namespace<Namespace> >
{
        typedef BOOST_MIRROR_GET_GLOBAL_LIST_BASE(
                Namespace,
                BOOST_MIRROR_COUNTER_LUID,
                typename
        ) type;
};

namespace _namespace {

// This structure provides basic meta-data about the global scope namespace
/*  Do not use this structure directly, use the appropriate specialization
 *  of meta_namespace instead.
 *  .see meta_namespace
 */
struct _
{
        typedef boost::mirror::meta_namespace_tag category;

        // The scope of this namespace is again the global scope
        typedef _ scope;

        // Implement the individual overloads of the name function
        BOOST_MIRROR_IMPLEMENT_META_OBJECT_NAME_FUNCTIONS("")
};

// This typedef is used by the nested namespaces
typedef _ _parent_namespace;

} // namespace _namespace

// Initialize the global list of members of this namespace
BOOST_MIRROR_REGISTER_GLOBAL_LIST( boost::mirror::_namespace::_ )

namespace _namespace {

// This structure provides basic meta-data about unspecified local scopes
/*  Do not use this structure directly, use the appropriate specialization
 *  of meta_namespace instead.
 *  .see meta_namespace
 */
struct _unspec_scope
{
        typedef boost::mirror::meta_unspecified_scope_tag category;

        // The scope of this namespace is again the global scope
        typedef _ scope;

        // Implement the individual overloads of the name function
        BOOST_MIRROR_IMPLEMENT_META_OBJECT_NAME_FUNCTIONS("<unspecified>")
};

} // namespace _namespace

// This macro implements the _ structure containing the basic metadata
#define BOOST_MIRROR_REGISTER_BASE_NAMESPACE_META_DATA(NAME) \
struct _ \
{  \
        typedef meta_namespace_tag category; \
        typedef _parent_namespace scope; \
        BOOST_MIRROR_IMPLEMENT_META_OBJECT_NAME_FUNCTIONS(#NAME) \
}; \
typedef _ _parent_namespace;

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// This macro begins the registering of namespace(s) with Mirror
/**
 *  @see BOOST_MIRROR_REG_NAMESPACE_END
 *  @see BOOST_MIRROR_REG_NAMESPACE
 *  @ingroup registering_macros
 */
#define BOOST_MIRROR_REG_NAMESPACE_BEGIN
#else
#define BOOST_MIRROR_REG_NAMESPACE_BEGIN namespace _namespace {
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// This macro registers a namespace with Mirror
/**
 *  @see BOOST_MIRROR_REG_NAMESPACE_BEGIN
 *  @see BOOST_MIRROR_REG_NAMESPACE_END
 *  @see BOOST_MIRRORED_NAMESPACE
 *  @ingroup registering_macros
 */
#define BOOST_MIRROR_REG_NAMESPACE(NAMESPACE_NAME)
#else
#define BOOST_MIRROR_REG_NAMESPACE(NAMESPACE_NAME) \
namespace NAMESPACE_NAME { \
BOOST_MIRROR_REGISTER_BASE_NAMESPACE_META_DATA(NAMESPACE_NAME) \
}
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// This macro ends the registering of namespace(s) with Mirror
/**
 *  @see BOOST_MIRROR_REG_NAMESPACE_BEGIN
 *  @see BOOST_MIRROR_REG_NAMESPACE
 *  @see BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST
 *  @ingroup registering_macros
 */
#define BOOST_MIRROR_REG_NAMESPACE_END
#else
#define BOOST_MIRROR_REG_NAMESPACE_END }
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// Macro which initializes the namespace's member list
/** This macro initializes the compile-time list of the members
 *  of the @a FULL_NAMESPACE_NAME namespace.
 *  This macro must be used between the @c BOOST_MIRROR_REG_BEGIN
 *  and the @c BOOST_MIRROR_REG_END macros.
 *  after registering the namespace, i.e. after the
 *  BOOST_MIRROR_REG_NAMESPACE_END macro.
 *
 *  @see BOOST_MIRROR_REG_NAMESPACE_BEGIN
 *  @see BOOST_MIRROR_REG_NAMESPACE
 *  @see BOOST_MIRROR_REG_NAMESPACE_END
 *  @ingroup registering_macros
 */
#define BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST(FULL_NAMESPACE_NAME)
#else
#define BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST(FULL_NAMESPACE_NAME) \
BOOST_MIRROR_REGISTER_GLOBAL_LIST( \
        boost::mirror::_namespace:: FULL_NAMESPACE_NAME ::_ \
) \
BOOST_MIRROR_ADD_TO_GLOBAL_LIST( \
        boost::mirror::_namespace:: FULL_NAMESPACE_NAME ::_::scope, \
        boost::mirror::meta_namespace< \
                boost::mirror::_namespace:: FULL_NAMESPACE_NAME ::_ \
         > \
)
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// Macro for quick registering of a single namespace under the global scope
/** This macro can be used for registering of a single namespace defined
 *  directly in the global scope i.e. not a nested namespace. This macro
 *  takes care of registering the namespace and initializing the member type
 *  list.
 *
 *  @param NAMESPACE the name of the namespace to be registered
 *  @see BOOST_MIRROR_REG_NAMESPACE_BEGIN
 *  @see BOOST_MIRROR_REG_NAMESPACE
 *  @see BOOST_MIRROR_REG_NAMESPACE_END
 *  @see BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST
 *  @see BOOST_MIRROR_QREG_NESTED_NAMESPACE
 *  @ingroup quick_registering_macros
 */
#define BOOST_MIRROR_QREG_GLOBAL_SCOPE_NAMESPACE(NAMESPACE)
#else
#define BOOST_MIRROR_QREG_GLOBAL_SCOPE_NAMESPACE(NAMESPACE) \
BOOST_MIRROR_REG_NAMESPACE_BEGIN \
BOOST_MIRROR_REG_NAMESPACE(NAMESPACE) \
BOOST_MIRROR_REG_NAMESPACE_END \
BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST(NAMESPACE)
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// Macro for quick registering of a namespace nested in top-level namespace
/** This macro can be used for registering of a single namespace defined
 *  in another namespace nested directly in the global scope. This macro
 *  takes care of registering the namespace and initializing the member type
 *  list. The enclosing parent namespace must be registered before the
 *  nested namespace is registered.
 *
 *  @param PARENT_NAMESPACE the name of the parent namespace
 *  @param NESTED_NAMESPACE the name of the nested namespace to be registered

 *  @see BOOST_MIRROR_REG_NAMESPACE_BEGIN
 *  @see BOOST_MIRROR_REG_NAMESPACE
 *  @see BOOST_MIRROR_REG_NAMESPACE_END
 *  @see BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST
 *  @see BOOST_MIRROR_QREG_GLOBAL_SCOPE_NAMESPACE
 *  @ingroup quick_registering_macros
 */
#define BOOST_MIRROR_QREG_NESTED_NAMESPACE(PARENT_NAMESPACE, NESTED_NAMESPACE)
#else
#define BOOST_MIRROR_QREG_NESTED_NAMESPACE(PARENT_NAMESPACE, NESTED_NAMESPACE)\
BOOST_MIRROR_REG_NAMESPACE_BEGIN \
namespace PARENT_NAMESPACE { \
BOOST_MIRROR_REG_NAMESPACE(NESTED_NAMESPACE) \
} \
BOOST_MIRROR_REG_NAMESPACE_END \
BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST(PARENT_NAMESPACE::NESTED_NAMESPACE)
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// Macro for quick registering of a namespace nested in a nested namespace
/** This macro can be used for registering of a single namespace defined
 *  in another namespace nested in namespace under the global scope. This macro
 *  takes care of registering the namespace and initializing the member type
 *  list. The enclosing parent namespace must be registered before the
 *  nested namespace is registered.
 *
 *  @param GRANDPARENT_NAMESPACE the name of the top-level namespace
 *  @param PARENT_NAMESPACE the name of the parent namespace
 *  @param NESTED_NAMESPACE the name of the nested namespace to be registered

 *  @see BOOST_MIRROR_REG_NAMESPACE_BEGIN
 *  @see BOOST_MIRROR_REG_NAMESPACE
 *  @see BOOST_MIRROR_REG_NAMESPACE_END
 *  @see BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST
 *  @see BOOST_MIRROR_QREG_GLOBAL_SCOPE_NAMESPACE
 *  @see BOOST_MIRROR_QREG_NESTED_NAMESPACE
 *  @ingroup quick_registering_macros
 */
#define BOOST_MIRROR_QREG_NESTED_NESTED_NAMESPACE( \
        GRANDPARENT_NAMESPACE, \
        PARENT_NAMESPACE, \
        NESTED_NAMESPACE \
)
#else
#define BOOST_MIRROR_QREG_NESTED_NESTED_NAMESPACE( \
        GRANDPARENT_NAMESPACE, \
        PARENT_NAMESPACE, \
        NESTED_NAMESPACE \
)\
BOOST_MIRROR_REG_NAMESPACE_BEGIN \
namespace GRANDPARENT_NAMESPACE { \
namespace PARENT_NAMESPACE { \
BOOST_MIRROR_REG_NAMESPACE(NESTED_NAMESPACE) \
} } \
BOOST_MIRROR_REG_NAMESPACE_END \
BOOST_MIRROR_INIT_NAMESPACE_MEMBER_LIST( \
        GRANDPARENT_NAMESPACE::PARENT_NAMESPACE::NESTED_NAMESPACE \
)
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// Reflects the namespace passed as the @a FULL_NAMESPACE_NAME argument
/** @def BOOST_MIRRORED_NAMESPACE(FULL_NAMESPACE_NAME)
 *  This macro expands into a type conforming to the MetaNamespace
 *  concept, which provides meta-data about the given namespace.
 *  The namespace name passed as the @a FULL_NAMESPACE_NAME argument
 *  must NOT contain the leading :: (double colon).
 *
 *  @see BOOST_MIRRORED_GLOBAL_SCOPE
 *  @see BOOST_MIRROR_REG_NAMESPACE
 *  @see boost::mirror::MetaNamespace
 *
 *  @ingroup reflection_macros
 */
#define BOOST_MIRRORED_NAMESPACE(FULL_NAMESPACE_NAME) boost::mirror::MetaNamespace
#else
#define BOOST_MIRRORED_NAMESPACE(FULL_NAMESPACE_NAME) \
        boost::mirror::meta_namespace< \
        boost::mirror::_namespace :: FULL_NAMESPACE_NAME :: _ \
        >
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// Reflects the global scope namespace
/** @def BOOST_MIRRORED_GLOBAL_SCOPE
 *  This macro expands into a type conforming to the MetaNamespace
 *  concept, which provides meta-data about the given namespace.
 *
 *  @see BOOST_MIRRORED_NAMESPACE
 *  @see boost::mirror::MetaNamespace
 *  @see boost::lagoon::reflected_global_scope
 *
 *  @ingroup reflection_macros
 */
#define BOOST_MIRRORED_GLOBAL_SCOPE() boost::mirror::MetaNamespace
#else
#define BOOST_MIRRORED_GLOBAL_SCOPE() \
boost::mirror::meta_namespace< \
        boost::mirror::_namespace:: _ \
>
#endif

#ifdef BOOST_MIRROR_DOCUMENTATION_ONLY
/// Reflects an unspecified (local) scope
/** @def BOOST_MIRRORED_UNSPECIFIED_SCOPE
 *  This macro expands into a type conforming to the MetaNamespace
 *  concept, which provides meta-data about an unspecified scope
 *  used in cases where the scope of an MetaScopedObject is not
 *  important.
 *
 *  @see BOOST_MIRRORED_NAMESPACE
 *  @see boost::mirror::MetaNamespace
 *
 *  @ingroup reflection_macros
 */
#define BOOST_MIRRORED_UNSPECIFIED_SCOPE() boost::mirror::MetaNamespace
#else
#define BOOST_MIRRORED_UNSPECIFIED_SCOPE() \
boost::mirror::meta_namespace< \
        boost::mirror::_namespace:: _unspec_scope \
>
#endif


BOOST_MIRROR_NAMESPACE_END

#endif //include guard

